@@ -420,8 +420,9 @@ simple_bind_pair env@(SOE { soe_inl = inl_env, soe_subst = subst })
     safe_to_inline :: OccInfo -> Bool
     safe_to_inline (IAmALoopBreaker {}) = False
     safe_to_inline IAmDead              = True
-    safe_to_inline occ@(OneOcc {})      =  not (occ_in_lam occ)
-                                        && occ_one_br occ
+    safe_to_inline OneOcc{ occ_in_lam = False
+                         , occ_n_br = 1 }             = True
+    safe_to_inline OneOcc{}                           = False
     safe_to_inline (ManyOccs {})        = False
 
 -------------------
