@@ -34,7 +34,7 @@ module TcHsSyn (
         zonkTopBndrs,
         ZonkEnv, ZonkFlexi(..), emptyZonkEnv, mkEmptyZonkEnv, initZonkEnv,
         zonkTyVarBinders, zonkTyVarBindersX, zonkTyVarBinderX,
-        zonkTyBndrs, zonkTyBndrsX, zonkRecTyVarBndrs,
+        zonkTyBndrs, zonkTyBndrsX,
         zonkTcTypeToType,  zonkTcTypeToTypeX,
         zonkTcTypesToTypes, zonkTcTypesToTypesX,
         zonkTyVarOcc,
@@ -325,20 +325,14 @@ extendZonkEnv ze@(ZonkEnv { ze_tv_env = tyco_env, ze_id_env = id_env }) vars
   where
     (tycovars, ids) = partition isTyCoVar vars
 
-extendIdZonkEnv1 :: ZonkEnv -> Var -> ZonkEnv
-extendIdZonkEnv1 ze@(ZonkEnv { ze_id_env = id_env }) id
+extendIdZonkEnv :: ZonkEnv -> Var -> ZonkEnv
+extendIdZonkEnv ze@(ZonkEnv { ze_id_env = id_env }) id
   = ze { ze_id_env = extendVarEnv id_env id id }
 
-extendTyZonkEnv1 :: ZonkEnv -> TyVar -> ZonkEnv
-extendTyZonkEnv1 ze@(ZonkEnv { ze_tv_env = ty_env }) tv
+extendTyZonkEnv :: ZonkEnv -> TyVar -> ZonkEnv
+extendTyZonkEnv ze@(ZonkEnv { ze_tv_env = ty_env }) tv
   = ze { ze_tv_env = extendVarEnv ty_env tv tv }
 
-extendTyZonkEnvN :: ZonkEnv -> [(Name,TyVar)] -> ZonkEnv
-extendTyZonkEnvN ze@(ZonkEnv { ze_tv_env = ty_env }) pairs
-  = ze { ze_tv_env = foldl add ty_env pairs }
-  where
-    add env (name, tv) = extendVarEnv_Directly env (getUnique name) tv
-
 setZonkType :: ZonkEnv -> ZonkFlexi -> ZonkEnv
 setZonkType ze flexi = ze { ze_flexi = flexi }
 
@@ -427,7 +421,7 @@ zonkEvVarOcc env v
 zonkCoreBndrX :: ZonkEnv -> Var -> TcM (ZonkEnv, Var)
 zonkCoreBndrX env v
   | isId v = do { v' <- zonkIdBndr env v
-                ; return (extendIdZonkEnv1 env v', v') }
+                ; return (extendIdZonkEnv env v', v') }
   | otherwise = zonkTyBndrX env v
 
 zonkCoreBndrsX :: ZonkEnv -> [Var] -> TcM (ZonkEnv, [Var])
@@ -442,12 +436,16 @@ zonkTyBndrsX = mapAccumLM zonkTyBndrX
 zonkTyBndrX :: ZonkEnv -> TcTyVar -> TcM (ZonkEnv, TyVar)
 -- This guarantees to return a TyVar (not a TcTyVar)
 -- then we add it to the envt, so all occurrences are replaced
+--
+-- It does not clone: the new TyVar has the sane Name
+-- as the old one.  This important when zonking the
+-- TyVarBndrs of a TyCon, whose Names may scope.
 zonkTyBndrX env tv
   = ASSERT2( isImmutableTyVar tv, ppr tv <+> dcolon <+> ppr (tyVarKind tv) )
     do { ki <- zonkTcTypeToTypeX env (tyVarKind tv)
                -- Internal names tidy up better, for iface files.
        ; let tv' = mkTyVar (tyVarName tv) ki
-       ; return (extendTyZonkEnv1 env tv', tv') }
+       ; return (extendTyZonkEnv env tv', tv') }
 
 zonkTyVarBinders ::  [VarBndr TcTyVar vis]
                  -> TcM (ZonkEnv, [VarBndr TyVar vis])
@@ -464,22 +462,6 @@ zonkTyVarBinderX env (Bndr tv vis)
   = do { (env', tv') <- zonkTyBndrX env tv
        ; return (env', Bndr tv' vis) }
 
-zonkRecTyVarBndrs :: [Name] -> [TcTyVar] -> TcM (ZonkEnv, [TyVar])
--- This rather specialised function is used in exactly one place.
--- See Note [Tricky scoping in generaliseTcTyCon] in TcTyClsDecls.
-zonkRecTyVarBndrs names tc_tvs
-  = initZonkEnv $ \ ze ->
-    fixM $ \ ~(_, rec_new_tvs) ->
-    do { let ze' = extendTyZonkEnvN ze $
-                   zipWithLazy (\ tc_tv new_tv -> (getName tc_tv, new_tv))
-                               tc_tvs rec_new_tvs
-       ; new_tvs <- zipWithM (zonk_one ze') names tc_tvs
-       ; return (ze', new_tvs) }
-  where
-    zonk_one ze name tc_tv
-      = do { ki <- zonkTcTypeToTypeX ze (tyVarKind tc_tv)
-           ; return (mkTyVar name ki) }
-
 zonkTopExpr :: HsExpr GhcTcId -> TcM (HsExpr GhcTc)
 zonkTopExpr e = initZonkEnv $ \ ze -> zonkExpr ze e
 
@@ -1361,7 +1343,7 @@ zonk_pat env (WildPat ty)
 
 zonk_pat env (VarPat x (dL->L l v))
   = do  { v' <- zonkIdBndr env v
-        ; return (extendIdZonkEnv1 env v', VarPat x (cL l v')) }
+        ; return (extendIdZonkEnv env v', VarPat x (cL l v')) }
 
 zonk_pat env (LazyPat x pat)
   = do  { (env', pat') <- zonkPat env pat
@@ -1373,7 +1355,7 @@ zonk_pat env (BangPat x pat)
 
 zonk_pat env (AsPat x (dL->L loc v) pat)
   = do  { v' <- zonkIdBndr env v
-        ; (env', pat') <- zonkPat (extendIdZonkEnv1 env v') pat
+        ; (env', pat') <- zonkPat (extendIdZonkEnv env v') pat
         ; return (env', AsPat x (cL loc v') pat') }
 
 zonk_pat env (ViewPat ty expr pat)
@@ -1463,7 +1445,7 @@ zonk_pat env (NPlusKPat ty (dL->L loc n) (dL->L l lit1) lit2 e1 e2)
         ; lit1' <- zonkOverLit env2 lit1
         ; lit2' <- zonkOverLit env2 lit2
         ; ty' <- zonkTcTypeToTypeX env2 ty
-        ; return (extendIdZonkEnv1 env2 n',
+        ; return (extendIdZonkEnv env2 n',
                   NPlusKPat ty' (cL loc n') (cL l lit1') lit2' e1' e2') }
 
 zonk_pat env (CoPat x co_fn pat ty)
@@ -1613,7 +1595,7 @@ zonkCoreExpr env (Case scrut b ty alts)
     = do scrut' <- zonkCoreExpr env scrut
          ty' <- zonkTcTypeToTypeX env ty
          b' <- zonkIdBndr env b
-         let env1 = extendIdZonkEnv1 env b'
+         let env1 = extendIdZonkEnv env b'
          alts' <- mapM (zonkCoreAlt env1) alts
          return $ Case scrut' b' ty' alts'
 
@@ -1627,7 +1609,7 @@ zonkCoreBind :: ZonkEnv -> CoreBind -> TcM (ZonkEnv, CoreBind)
 zonkCoreBind env (NonRec v e)
     = do v' <- zonkIdBndr env v
          e' <- zonkCoreExpr env e
-         let env1 = extendIdZonkEnv1 env v'
+         let env1 = extendIdZonkEnv env v'
          return (env1, NonRec v' e')
 zonkCoreBind env (Rec pairs)
     = do (env1, pairs') <- fixM go
