@@ -1,17 +1,16 @@
 
 import CmmExpr
-import FastBool
-#if !(MACHREGS_i386 || MACHREGS_x86_64 || MACHREGS_sparc || MACHREGS_powerpc)
-import Panic
+#if !(defined(MACHREGS_i386) || defined(MACHREGS_x86_64) \
+    || defined(MACHREGS_sparc) || defined(MACHREGS_powerpc))
+import PlainPanic
 #endif
 import Reg
 
-#include "ghcautoconf.h"
 #include "stg/MachRegs.h"
 
-#if MACHREGS_i386 || MACHREGS_x86_64
+#if defined(MACHREGS_i386) || defined(MACHREGS_x86_64)
 
-# if MACHREGS_i386
+# if defined(MACHREGS_i386)
 #  define eax 0
 #  define ebx 1
 #  define ecx 2
@@ -22,7 +21,7 @@ import Reg
 #  define esp 7
 # endif
 
-# if MACHREGS_x86_64
+# if defined(MACHREGS_x86_64)
 #  define rax   0
 #  define rbx   1
 #  define rcx   2
@@ -41,70 +40,67 @@ import Reg
 #  define r15   15
 # endif
 
-# define fake0 16
-# define fake1 17
-# define fake2 18
-# define fake3 19
-# define fake4 20
-# define fake5 21
 
-# define xmm0  24
-# define xmm1  25
-# define xmm2  26
-# define xmm3  27
-# define xmm4  28
-# define xmm5  29
-# define xmm6  30
-# define xmm7  31
-# define xmm8  32
-# define xmm9  33
-# define xmm10 34
-# define xmm11 35
-# define xmm12 36
-# define xmm13 37
-# define xmm14 38
-# define xmm15 39
+-- N.B. XMM, YMM, and ZMM are all aliased to the same hardware registers hence
+-- being assigned the same RegNos.
+# define xmm0  16
+# define xmm1  17
+# define xmm2  18
+# define xmm3  19
+# define xmm4  20
+# define xmm5  21
+# define xmm6  22
+# define xmm7  23
+# define xmm8  24
+# define xmm9  25
+# define xmm10 26
+# define xmm11 27
+# define xmm12 28
+# define xmm13 29
+# define xmm14 30
+# define xmm15 31
 
-# define ymm0  40
-# define ymm1  41
-# define ymm2  42
-# define ymm3  43
-# define ymm4  44
-# define ymm5  45
-# define ymm6  46
-# define ymm7  47
-# define ymm8  48
-# define ymm9  49
-# define ymm10 50
-# define ymm11 51
-# define ymm12 52
-# define ymm13 53
-# define ymm14 54
-# define ymm15 55
+# define ymm0  16
+# define ymm1  17
+# define ymm2  18
+# define ymm3  19
+# define ymm4  20
+# define ymm5  21
+# define ymm6  22
+# define ymm7  23
+# define ymm8  24
+# define ymm9  25
+# define ymm10 26
+# define ymm11 27
+# define ymm12 28
+# define ymm13 29
+# define ymm14 30
+# define ymm15 31
 
-# define zmm0  56
-# define zmm1  57
-# define zmm2  58
-# define zmm3  59
-# define zmm4  60
-# define zmm5  61
-# define zmm6  62
-# define zmm7  63
-# define zmm8  64
-# define zmm9  65
-# define zmm10 66
-# define zmm11 67
-# define zmm12 68
-# define zmm13 69
-# define zmm14 70
-# define zmm15 71
+# define zmm0  16
+# define zmm1  17
+# define zmm2  18
+# define zmm3  19
+# define zmm4  20
+# define zmm5  21
+# define zmm6  22
+# define zmm7  23
+# define zmm8  24
+# define zmm9  25
+# define zmm10 26
+# define zmm11 27
+# define zmm12 28
+# define zmm13 29
+# define zmm14 30
+# define zmm15 31
 
--- Note: these are only needed for ARM/ARM64 because globalRegMaybe is now used in CmmSink.hs.
+-- Note: these are only needed for ARM/AArch64 because globalRegMaybe is now used in CmmSink.hs.
 -- Since it's only used to check 'isJust', the actual values don't matter, thus
 -- I'm not sure if these are the correct numberings.
 -- Normally, the register names are just stringified as part of the REG() macro
 
-#elif MACHREGS_powerpc || MACHREGS_arm || MACHREGS_aarch64
+#elif defined(MACHREGS_powerpc) || defined(MACHREGS_arm) \
+    || defined(MACHREGS_aarch64)
 
 # define r0 0
 # define r1 1
@@ -140,8 +136,8 @@ import Reg
 # define r31 31
 
 -- See note above. These aren't actually used for anything except satisfying the compiler for globalRegMaybe
--- so I'm unsure if they're the correct numberings, should they ever be attempted to be used in the NCG. 
-#if MACHREGS_aarch64 || MACHREGS_arm
+-- so I'm unsure if they're the correct numberings, should they ever be attempted to be used in the NCG.
+#if defined(MACHREGS_aarch64) || defined(MACHREGS_arm)
 # define s0 32
 # define s1 33
 # define s2 34
@@ -209,7 +205,7 @@ import Reg
 # define d31 63
 #endif
 
-# if MACHREGS_darwin
+# if defined(MACHREGS_darwin)
 #  define f0  32
 #  define f1  33
 #  define f2  34
@@ -277,7 +273,7 @@ import Reg
 #  define fr31 63
 # endif
 
-#elif MACHREGS_sparc
+#elif defined(MACHREGS_sparc)
 
 # define g0  0
 # define g1  1
@@ -348,281 +344,317 @@ import Reg
 # define f30 62
 # define f31 63
 
+#elif defined(MACHREGS_s390x)
+
+# define r0   0
+# define r1   1
+# define r2   2
+# define r3   3
+# define r4   4
+# define r5   5
+# define r6   6
+# define r7   7
+# define r8   8
+# define r9   9
+# define r10 10
+# define r11 11
+# define r12 12
+# define r13 13
+# define r14 14
+# define r15 15
+
+# define f0  16
+# define f1  17
+# define f2  18
+# define f3  19
+# define f4  20
+# define f5  21
+# define f6  22
+# define f7  23
+# define f8  24
+# define f9  25
+# define f10 26
+# define f11 27
+# define f12 28
+# define f13 29
+# define f14 30
+# define f15 31
+
 #endif
 
 callerSaves :: GlobalReg -> Bool
-#ifdef CALLER_SAVES_Base
+#if defined(CALLER_SAVES_Base)
 callerSaves BaseReg           = True
 #endif
-#ifdef CALLER_SAVES_R1
+#if defined(CALLER_SAVES_R1)
 callerSaves (VanillaReg 1 _)  = True
 #endif
-#ifdef CALLER_SAVES_R2
+#if defined(CALLER_SAVES_R2)
 callerSaves (VanillaReg 2 _)  = True
 #endif
-#ifdef CALLER_SAVES_R3
+#if defined(CALLER_SAVES_R3)
 callerSaves (VanillaReg 3 _)  = True
 #endif
-#ifdef CALLER_SAVES_R4
+#if defined(CALLER_SAVES_R4)
 callerSaves (VanillaReg 4 _)  = True
 #endif
-#ifdef CALLER_SAVES_R5
+#if defined(CALLER_SAVES_R5)
 callerSaves (VanillaReg 5 _)  = True
 #endif
-#ifdef CALLER_SAVES_R6
+#if defined(CALLER_SAVES_R6)
 callerSaves (VanillaReg 6 _)  = True
 #endif
-#ifdef CALLER_SAVES_R7
+#if defined(CALLER_SAVES_R7)
 callerSaves (VanillaReg 7 _)  = True
 #endif
-#ifdef CALLER_SAVES_R8
+#if defined(CALLER_SAVES_R8)
 callerSaves (VanillaReg 8 _)  = True
 #endif
-#ifdef CALLER_SAVES_R9
+#if defined(CALLER_SAVES_R9)
 callerSaves (VanillaReg 9 _)  = True
 #endif
-#ifdef CALLER_SAVES_R10
+#if defined(CALLER_SAVES_R10)
 callerSaves (VanillaReg 10 _) = True
 #endif
-#ifdef CALLER_SAVES_F1
+#if defined(CALLER_SAVES_F1)
 callerSaves (FloatReg 1)      = True
 #endif
-#ifdef CALLER_SAVES_F2
+#if defined(CALLER_SAVES_F2)
 callerSaves (FloatReg 2)      = True
 #endif
-#ifdef CALLER_SAVES_F3
+#if defined(CALLER_SAVES_F3)
 callerSaves (FloatReg 3)      = True
 #endif
-#ifdef CALLER_SAVES_F4
+#if defined(CALLER_SAVES_F4)
 callerSaves (FloatReg 4)      = True
 #endif
-#ifdef CALLER_SAVES_F5
+#if defined(CALLER_SAVES_F5)
 callerSaves (FloatReg 5)      = True
 #endif
-#ifdef CALLER_SAVES_F6
+#if defined(CALLER_SAVES_F6)
 callerSaves (FloatReg 6)      = True
 #endif
-#ifdef CALLER_SAVES_D1
+#if defined(CALLER_SAVES_D1)
 callerSaves (DoubleReg 1)     = True
 #endif
-#ifdef CALLER_SAVES_D2
+#if defined(CALLER_SAVES_D2)
 callerSaves (DoubleReg 2)     = True
 #endif
-#ifdef CALLER_SAVES_D3
+#if defined(CALLER_SAVES_D3)
 callerSaves (DoubleReg 3)     = True
 #endif
-#ifdef CALLER_SAVES_D4
+#if defined(CALLER_SAVES_D4)
 callerSaves (DoubleReg 4)     = True
 #endif
-#ifdef CALLER_SAVES_D5
+#if defined(CALLER_SAVES_D5)
 callerSaves (DoubleReg 5)     = True
 #endif
-#ifdef CALLER_SAVES_D6
+#if defined(CALLER_SAVES_D6)
 callerSaves (DoubleReg 6)     = True
 #endif
-#ifdef CALLER_SAVES_L1
+#if defined(CALLER_SAVES_L1)
 callerSaves (LongReg 1)       = True
 #endif
-#ifdef CALLER_SAVES_Sp
+#if defined(CALLER_SAVES_Sp)
 callerSaves Sp                = True
 #endif
-#ifdef CALLER_SAVES_SpLim
+#if defined(CALLER_SAVES_SpLim)
 callerSaves SpLim             = True
 #endif
-#ifdef CALLER_SAVES_Hp
+#if defined(CALLER_SAVES_Hp)
 callerSaves Hp                = True
 #endif
-#ifdef CALLER_SAVES_HpLim
+#if defined(CALLER_SAVES_HpLim)
 callerSaves HpLim             = True
 #endif
-#ifdef CALLER_SAVES_CCCS
+#if defined(CALLER_SAVES_CCCS)
 callerSaves CCCS              = True
 #endif
-#ifdef CALLER_SAVES_CurrentTSO
+#if defined(CALLER_SAVES_CurrentTSO)
 callerSaves CurrentTSO        = True
 #endif
-#ifdef CALLER_SAVES_CurrentNursery
+#if defined(CALLER_SAVES_CurrentNursery)
 callerSaves CurrentNursery    = True
 #endif
 callerSaves _                 = False
 
 activeStgRegs :: [GlobalReg]
 activeStgRegs = [
-#ifdef REG_Base
+#if defined(REG_Base)
     BaseReg
 #endif
-#ifdef REG_Sp
+#if defined(REG_Sp)
     ,Sp
 #endif
-#ifdef REG_Hp
+#if defined(REG_Hp)
     ,Hp
 #endif
-#ifdef REG_R1
+#if defined(REG_R1)
     ,VanillaReg 1 VGcPtr
 #endif
-#ifdef REG_R2
+#if defined(REG_R2)
     ,VanillaReg 2 VGcPtr
 #endif
-#ifdef REG_R3
+#if defined(REG_R3)
     ,VanillaReg 3 VGcPtr
 #endif
-#ifdef REG_R4
+#if defined(REG_R4)
     ,VanillaReg 4 VGcPtr
 #endif
-#ifdef REG_R5
+#if defined(REG_R5)
     ,VanillaReg 5 VGcPtr
 #endif
-#ifdef REG_R6
+#if defined(REG_R6)
     ,VanillaReg 6 VGcPtr
 #endif
-#ifdef REG_R7
+#if defined(REG_R7)
     ,VanillaReg 7 VGcPtr
 #endif
-#ifdef REG_R8
+#if defined(REG_R8)
     ,VanillaReg 8 VGcPtr
 #endif
-#ifdef REG_R9
+#if defined(REG_R9)
     ,VanillaReg 9 VGcPtr
 #endif
-#ifdef REG_R10
+#if defined(REG_R10)
     ,VanillaReg 10 VGcPtr
 #endif
-#ifdef REG_SpLim
+#if defined(REG_SpLim)
     ,SpLim
 #endif
 #if MAX_REAL_XMM_REG != 0
-#ifdef REG_F1
+#if defined(REG_F1)
     ,FloatReg 1
 #endif
-#ifdef REG_D1
+#if defined(REG_D1)
     ,DoubleReg 1
 #endif
-#ifdef REG_XMM1
+#if defined(REG_XMM1)
     ,XmmReg 1
 #endif
-#ifdef REG_YMM1
+#if defined(REG_YMM1)
     ,YmmReg 1
 #endif
-#ifdef REG_ZMM1
+#if defined(REG_ZMM1)
     ,ZmmReg 1
 #endif
-#ifdef REG_F2
+#if defined(REG_F2)
     ,FloatReg 2
 #endif
-#ifdef REG_D2
+#if defined(REG_D2)
     ,DoubleReg 2
 #endif
-#ifdef REG_XMM2
+#if defined(REG_XMM2)
     ,XmmReg 2
 #endif
-#ifdef REG_YMM2
+#if defined(REG_YMM2)
     ,YmmReg 2
 #endif
-#ifdef REG_ZMM2
+#if defined(REG_ZMM2)
     ,ZmmReg 2
 #endif
-#ifdef REG_F3
+#if defined(REG_F3)
     ,FloatReg 3
 #endif
-#ifdef REG_D3
+#if defined(REG_D3)
     ,DoubleReg 3
 #endif
-#ifdef REG_XMM3
+#if defined(REG_XMM3)
     ,XmmReg 3
 #endif
-#ifdef REG_YMM3
+#if defined(REG_YMM3)
     ,YmmReg 3
 #endif
-#ifdef REG_ZMM3
+#if defined(REG_ZMM3)
     ,ZmmReg 3
 #endif
-#ifdef REG_F4
+#if defined(REG_F4)
     ,FloatReg 4
 #endif
-#ifdef REG_D4
+#if defined(REG_D4)
     ,DoubleReg 4
 #endif
-#ifdef REG_XMM4
+#if defined(REG_XMM4)
     ,XmmReg 4
 #endif
-#ifdef REG_YMM4
+#if defined(REG_YMM4)
     ,YmmReg 4
 #endif
-#ifdef REG_ZMM4
+#if defined(REG_ZMM4)
     ,ZmmReg 4
 #endif
-#ifdef REG_F5
+#if defined(REG_F5)
     ,FloatReg 5
 #endif
-#ifdef REG_D5
+#if defined(REG_D5)
     ,DoubleReg 5
 #endif
-#ifdef REG_XMM5
+#if defined(REG_XMM5)
     ,XmmReg 5
 #endif
-#ifdef REG_YMM5
+#if defined(REG_YMM5)
     ,YmmReg 5
 #endif
-#ifdef REG_ZMM5
+#if defined(REG_ZMM5)
     ,ZmmReg 5
 #endif
-#ifdef REG_F6
+#if defined(REG_F6)
     ,FloatReg 6
 #endif
-#ifdef REG_D6
+#if defined(REG_D6)
     ,DoubleReg 6
 #endif
-#ifdef REG_XMM6
+#if defined(REG_XMM6)
     ,XmmReg 6
 #endif
-#ifdef REG_YMM6
+#if defined(REG_YMM6)
     ,YmmReg 6
 #endif
-#ifdef REG_ZMM6
+#if defined(REG_ZMM6)
     ,ZmmReg 6
 #endif
 #else /* MAX_REAL_XMM_REG == 0 */
-#ifdef REG_F1
+#if defined(REG_F1)
     ,FloatReg 1
 #endif
-#ifdef REG_F2
+#if defined(REG_F2)
     ,FloatReg 2
 #endif
-#ifdef REG_F3
+#if defined(REG_F3)
     ,FloatReg 3
 #endif
-#ifdef REG_F4
+#if defined(REG_F4)
     ,FloatReg 4
 #endif
-#ifdef REG_F5
+#if defined(REG_F5)
     ,FloatReg 5
 #endif
-#ifdef REG_F6
+#if defined(REG_F6)
     ,FloatReg 6
 #endif
-#ifdef REG_D1
+#if defined(REG_D1)
     ,DoubleReg 1
 #endif
-#ifdef REG_D2
+#if defined(REG_D2)
     ,DoubleReg 2
 #endif
-#ifdef REG_D3
+#if defined(REG_D3)
     ,DoubleReg 3
 #endif
-#ifdef REG_D4
+#if defined(REG_D4)
     ,DoubleReg 4
 #endif
-#ifdef REG_D5
+#if defined(REG_D5)
     ,DoubleReg 5
 #endif
-#ifdef REG_D6
+#if defined(REG_D6)
     ,DoubleReg 6
 #endif
 #endif /* MAX_REAL_XMM_REG == 0 */
     ]
 
 haveRegBase :: Bool
-#ifdef REG_Base
+#if defined(REG_Base)
 haveRegBase = True
 #else
 haveRegBase = False
@@ -632,212 +664,219 @@ haveRegBase = False
 -- in a real machine register, otherwise returns @'Just' reg@, where
 -- reg is the machine register it is stored in.
 globalRegMaybe :: GlobalReg -> Maybe RealReg
-#if MACHREGS_i386 || MACHREGS_x86_64 || MACHREGS_sparc || MACHREGS_powerpc || MACHREGS_arm || MACHREGS_aarch64
-# ifdef REG_Base
+#if defined(MACHREGS_i386) || defined(MACHREGS_x86_64) \
+    || defined(MACHREGS_sparc) || defined(MACHREGS_powerpc) \
+    || defined(MACHREGS_arm) || defined(MACHREGS_aarch64) \
+    || defined(MACHREGS_s390x)
+# if defined(REG_Base)
 globalRegMaybe BaseReg                  = Just (RealRegSingle REG_Base)
 # endif
-# ifdef REG_R1
+# if defined(REG_R1)
 globalRegMaybe (VanillaReg 1 _)         = Just (RealRegSingle REG_R1)
 # endif
-# ifdef REG_R2
+# if defined(REG_R2)
 globalRegMaybe (VanillaReg 2 _)         = Just (RealRegSingle REG_R2)
 # endif
-# ifdef REG_R3
+# if defined(REG_R3)
 globalRegMaybe (VanillaReg 3 _)         = Just (RealRegSingle REG_R3)
 # endif
-# ifdef REG_R4
+# if defined(REG_R4)
 globalRegMaybe (VanillaReg 4 _)         = Just (RealRegSingle REG_R4)
 # endif
-# ifdef REG_R5
+# if defined(REG_R5)
 globalRegMaybe (VanillaReg 5 _)         = Just (RealRegSingle REG_R5)
 # endif
-# ifdef REG_R6
+# if defined(REG_R6)
 globalRegMaybe (VanillaReg 6 _)         = Just (RealRegSingle REG_R6)
 # endif
-# ifdef REG_R7
+# if defined(REG_R7)
 globalRegMaybe (VanillaReg 7 _)         = Just (RealRegSingle REG_R7)
 # endif
-# ifdef REG_R8
+# if defined(REG_R8)
 globalRegMaybe (VanillaReg 8 _)         = Just (RealRegSingle REG_R8)
 # endif
-# ifdef REG_R9
+# if defined(REG_R9)
 globalRegMaybe (VanillaReg 9 _)         = Just (RealRegSingle REG_R9)
 # endif
-# ifdef REG_R10
+# if defined(REG_R10)
 globalRegMaybe (VanillaReg 10 _)        = Just (RealRegSingle REG_R10)
 # endif
-# ifdef REG_F1
+# if defined(REG_F1)
 globalRegMaybe (FloatReg 1)             = Just (RealRegSingle REG_F1)
 # endif
-# ifdef REG_F2
+# if defined(REG_F2)
 globalRegMaybe (FloatReg 2)             = Just (RealRegSingle REG_F2)
 # endif
-# ifdef REG_F3
+# if defined(REG_F3)
 globalRegMaybe (FloatReg 3)             = Just (RealRegSingle REG_F3)
 # endif
-# ifdef REG_F4
+# if defined(REG_F4)
 globalRegMaybe (FloatReg 4)             = Just (RealRegSingle REG_F4)
 # endif
-# ifdef REG_F5
+# if defined(REG_F5)
 globalRegMaybe (FloatReg 5)             = Just (RealRegSingle REG_F5)
 # endif
-# ifdef REG_F6
+# if defined(REG_F6)
 globalRegMaybe (FloatReg 6)             = Just (RealRegSingle REG_F6)
 # endif
-# ifdef REG_D1
+# if defined(REG_D1)
 globalRegMaybe (DoubleReg 1)            =
-#  if MACHREGS_sparc
+#  if defined(MACHREGS_sparc)
                                           Just (RealRegPair REG_D1 (REG_D1 + 1))
 #  else
                                           Just (RealRegSingle REG_D1)
 #  endif
 # endif
-# ifdef REG_D2
+# if defined(REG_D2)
 globalRegMaybe (DoubleReg 2)            =
-#  if MACHREGS_sparc
+#  if defined(MACHREGS_sparc)
                                           Just (RealRegPair REG_D2 (REG_D2 + 1))
 #  else
                                           Just (RealRegSingle REG_D2)
 #  endif
 # endif
-# ifdef REG_D3
+# if defined(REG_D3)
 globalRegMaybe (DoubleReg 3)            =
-#  if MACHREGS_sparc
+#  if defined(MACHREGS_sparc)
                                           Just (RealRegPair REG_D3 (REG_D3 + 1))
 #  else
                                           Just (RealRegSingle REG_D3)
 #  endif
 # endif
-# ifdef REG_D4
+# if defined(REG_D4)
 globalRegMaybe (DoubleReg 4)            =
-#  if MACHREGS_sparc
+#  if defined(MACHREGS_sparc)
                                           Just (RealRegPair REG_D4 (REG_D4 + 1))
 #  else
                                           Just (RealRegSingle REG_D4)
 #  endif
 # endif
-# ifdef REG_D5
+# if defined(REG_D5)
 globalRegMaybe (DoubleReg 5)            =
-#  if MACHREGS_sparc
+#  if defined(MACHREGS_sparc)
                                           Just (RealRegPair REG_D5 (REG_D5 + 1))
 #  else
                                           Just (RealRegSingle REG_D5)
 #  endif
 # endif
-# ifdef REG_D6
+# if defined(REG_D6)
 globalRegMaybe (DoubleReg 6)            =
-#  if MACHREGS_sparc
+#  if defined(MACHREGS_sparc)
                                           Just (RealRegPair REG_D6 (REG_D6 + 1))
 #  else
                                           Just (RealRegSingle REG_D6)
 #  endif
 # endif
 # if MAX_REAL_XMM_REG != 0
-#  ifdef REG_XMM1
+#  if defined(REG_XMM1)
 globalRegMaybe (XmmReg 1)               = Just (RealRegSingle REG_XMM1)
 #  endif
-#  ifdef REG_XMM2
+#  if defined(REG_XMM2)
 globalRegMaybe (XmmReg 2)               = Just (RealRegSingle REG_XMM2)
 #  endif
-#  ifdef REG_XMM3
+#  if defined(REG_XMM3)
 globalRegMaybe (XmmReg 3)               = Just (RealRegSingle REG_XMM3)
 #  endif
-#  ifdef REG_XMM4
+#  if defined(REG_XMM4)
 globalRegMaybe (XmmReg 4)               = Just (RealRegSingle REG_XMM4)
 #  endif
-#  ifdef REG_XMM5
+#  if defined(REG_XMM5)
 globalRegMaybe (XmmReg 5)               = Just (RealRegSingle REG_XMM5)
 #  endif
-#  ifdef REG_XMM6
+#  if defined(REG_XMM6)
 globalRegMaybe (XmmReg 6)               = Just (RealRegSingle REG_XMM6)
 #  endif
 # endif
-# if MAX_REAL_YMM_REG != 0
-#  ifdef REG_YMM1
+# if defined(MAX_REAL_YMM_REG) && MAX_REAL_YMM_REG != 0
+#  if defined(REG_YMM1)
 globalRegMaybe (YmmReg 1)               = Just (RealRegSingle REG_YMM1)
 #  endif
-#  ifdef REG_YMM2
+#  if defined(REG_YMM2)
 globalRegMaybe (YmmReg 2)               = Just (RealRegSingle REG_YMM2)
 #  endif
-#  ifdef REG_YMM3
+#  if defined(REG_YMM3)
 globalRegMaybe (YmmReg 3)               = Just (RealRegSingle REG_YMM3)
 #  endif
-#  ifdef REG_YMM4
+#  if defined(REG_YMM4)
 globalRegMaybe (YmmReg 4)               = Just (RealRegSingle REG_YMM4)
 #  endif
-#  ifdef REG_YMM5
+#  if defined(REG_YMM5)
 globalRegMaybe (YmmReg 5)               = Just (RealRegSingle REG_YMM5)
 #  endif
-#  ifdef REG_YMM6
+#  if defined(REG_YMM6)
 globalRegMaybe (YmmReg 6)               = Just (RealRegSingle REG_YMM6)
 #  endif
 # endif
-# if MAX_REAL_ZMM_REG != 0
-#  ifdef REG_ZMM1
+# if defined(MAX_REAL_ZMM_REG) && MAX_REAL_ZMM_REG != 0
+#  if defined(REG_ZMM1)
 globalRegMaybe (ZmmReg 1)               = Just (RealRegSingle REG_ZMM1)
 #  endif
-#  ifdef REG_ZMM2
+#  if defined(REG_ZMM2)
 globalRegMaybe (ZmmReg 2)               = Just (RealRegSingle REG_ZMM2)
 #  endif
-#  ifdef REG_ZMM3
+#  if defined(REG_ZMM3)
 globalRegMaybe (ZmmReg 3)               = Just (RealRegSingle REG_ZMM3)
 #  endif
-#  ifdef REG_ZMM4
+#  if defined(REG_ZMM4)
 globalRegMaybe (ZmmReg 4)               = Just (RealRegSingle REG_ZMM4)
 #  endif
-#  ifdef REG_ZMM5
+#  if defined(REG_ZMM5)
 globalRegMaybe (ZmmReg 5)               = Just (RealRegSingle REG_ZMM5)
 #  endif
-#  ifdef REG_ZMM6
+#  if defined(REG_ZMM6)
 globalRegMaybe (ZmmReg 6)               = Just (RealRegSingle REG_ZMM6)
 #  endif
 # endif
-# ifdef REG_Sp
+# if defined(REG_Sp)
 globalRegMaybe Sp                       = Just (RealRegSingle REG_Sp)
 # endif
-# ifdef REG_Lng1
+# if defined(REG_Lng1)
 globalRegMaybe (LongReg 1)              = Just (RealRegSingle REG_Lng1)
 # endif
-# ifdef REG_Lng2
+# if defined(REG_Lng2)
 globalRegMaybe (LongReg 2)              = Just (RealRegSingle REG_Lng2)
 # endif
-# ifdef REG_SpLim
+# if defined(REG_SpLim)
 globalRegMaybe SpLim                    = Just (RealRegSingle REG_SpLim)
 # endif
-# ifdef REG_Hp
+# if defined(REG_Hp)
 globalRegMaybe Hp                       = Just (RealRegSingle REG_Hp)
 # endif
-# ifdef REG_HpLim
+# if defined(REG_HpLim)
 globalRegMaybe HpLim                    = Just (RealRegSingle REG_HpLim)
 # endif
-# ifdef REG_CurrentTSO
+# if defined(REG_CurrentTSO)
 globalRegMaybe CurrentTSO               = Just (RealRegSingle REG_CurrentTSO)
 # endif
-# ifdef REG_CurrentNursery
+# if defined(REG_CurrentNursery)
 globalRegMaybe CurrentNursery           = Just (RealRegSingle REG_CurrentNursery)
 # endif
+# if defined(REG_MachSp)
+globalRegMaybe MachSp                   = Just (RealRegSingle REG_MachSp)
+# endif
 globalRegMaybe _                        = Nothing
-#elif MACHREGS_NO_REGS
+#elif defined(MACHREGS_NO_REGS)
 globalRegMaybe _ = Nothing
 #else
 globalRegMaybe = panic "globalRegMaybe not defined for this platform"
 #endif
 
-freeReg :: RegNo -> FastBool
+freeReg :: RegNo -> Bool
 
-#if MACHREGS_i386 || MACHREGS_x86_64
+#if defined(MACHREGS_i386) || defined(MACHREGS_x86_64)
 
-# if MACHREGS_i386
-freeReg esp = fastBool False -- %esp is the C stack pointer
-freeReg esi = fastBool False -- Note [esi/edi not allocatable]
-freeReg edi = fastBool False
+# if defined(MACHREGS_i386)
+freeReg esp = False -- %esp is the C stack pointer
+freeReg esi = False -- Note [esi/edi/ebp not allocatable]
+freeReg edi = False
+freeReg ebp = False
 # endif
-# if MACHREGS_x86_64
-freeReg rsp = fastBool False  --        %rsp is the C stack pointer
+# if defined(MACHREGS_x86_64)
+freeReg rsp = False  --        %rsp is the C stack pointer
 # endif
 
 {-
-Note [esi/edi not allocatable]
+Note [esi/edi/ebp not allocatable]
 
 %esi is mapped to R1, so %esi would normally be allocatable while it
 is not being used for R1.  However, %esi has no 8-bit version on x86,
@@ -847,161 +886,100 @@ graph-colouring allocator also cannot handle this - it was designed
 with more flexibility in mind, but the current implementation is
 restricted to the same set of classes as the linear allocator.
 
-Hence, on x86 esi and edi are treated as not allocatable.
+Hence, on x86 esi, edi and ebp are treated as not allocatable.
 -}
 
 -- split patterns in two functions to prevent overlaps
 freeReg r         = freeRegBase r
 
-freeRegBase :: RegNo -> FastBool
-# ifdef REG_Base
-freeRegBase REG_Base = fastBool False
+freeRegBase :: RegNo -> Bool
+# if defined(REG_Base)
+freeRegBase REG_Base  = False
 # endif
-# ifdef REG_Sp
-freeRegBase REG_Sp   = fastBool False
+# if defined(REG_Sp)
+freeRegBase REG_Sp    = False
 # endif
-# ifdef REG_SpLim
-freeRegBase REG_SpLim = fastBool False
+# if defined(REG_SpLim)
+freeRegBase REG_SpLim = False
 # endif
-# ifdef REG_Hp
-freeRegBase REG_Hp   = fastBool False
+# if defined(REG_Hp)
+freeRegBase REG_Hp    = False
 # endif
-# ifdef REG_HpLim
-freeRegBase REG_HpLim = fastBool False
+# if defined(REG_HpLim)
+freeRegBase REG_HpLim = False
 # endif
 -- All other regs are considered to be "free", because we can track
 -- their liveness accurately.
-freeRegBase _ = fastBool True
+freeRegBase _ = True
 
-#elif MACHREGS_powerpc
+#elif defined(MACHREGS_powerpc)
 
-freeReg 0 = fastBool False -- Hack: r0 can't be used in all insns,
-                           -- but it's actually free
-freeReg 1 = fastBool False -- The Stack Pointer
-# if !MACHREGS_darwin
--- most non-darwin powerpc OSes use r2 as a TOC pointer or something like that
-freeReg 2 = fastBool False
+freeReg 0 = False -- Used by code setting the back chain pointer
+                  -- in stack reallocations on Linux.
+                  -- Moreover r0 is not usable in all insns.
+freeReg 1 = False -- The Stack Pointer
+-- most ELF PowerPC OSes use r2 as a TOC pointer
+freeReg 2 = False
+freeReg 13 = False -- reserved for system thread ID on 64 bit
 -- at least linux in -fPIC relies on r30 in PLT stubs
-freeReg 30 = fastBool False
-# endif
-# ifdef REG_Base
-freeReg REG_Base = fastBool False
-# endif
-# ifdef REG_R1
-freeReg REG_R1   = fastBool False
-# endif
-# ifdef REG_R2
-freeReg REG_R2   = fastBool False
-# endif
-# ifdef REG_R3
-freeReg REG_R3   = fastBool False
-# endif
-# ifdef REG_R4
-freeReg REG_R4   = fastBool False
-# endif
-# ifdef REG_R5
-freeReg REG_R5   = fastBool False
-# endif
-# ifdef REG_R6
-freeReg REG_R6   = fastBool False
-# endif
-# ifdef REG_R7
-freeReg REG_R7   = fastBool False
-# endif
-# ifdef REG_R8
-freeReg REG_R8   = fastBool False
-# endif
-# ifdef REG_R9
-freeReg REG_R9   = fastBool False
-# endif
-# ifdef REG_R10
-freeReg REG_R10  = fastBool False
-# endif
-# ifdef REG_F1
-freeReg REG_F1 = fastBool False
-# endif
-# ifdef REG_F2
-freeReg REG_F2 = fastBool False
-# endif
-# ifdef REG_F3
-freeReg REG_F3 = fastBool False
-# endif
-# ifdef REG_F4
-freeReg REG_F4 = fastBool False
-# endif
-# ifdef REG_F5
-freeReg REG_F5 = fastBool False
-# endif
-# ifdef REG_F6
-freeReg REG_F6 = fastBool False
-# endif
-# ifdef REG_D1
-freeReg REG_D1 = fastBool False
-# endif
-# ifdef REG_D2
-freeReg REG_D2 = fastBool False
-# endif
-# ifdef REG_D3
-freeReg REG_D3 = fastBool False
-# endif
-# ifdef REG_D4
-freeReg REG_D4 = fastBool False
-# endif
-# ifdef REG_D5
-freeReg REG_D5 = fastBool False
-# endif
-# ifdef REG_D6
-freeReg REG_D6 = fastBool False
-# endif
-# ifdef REG_Sp
-freeReg REG_Sp   = fastBool False
-# endif
-# ifdef REG_Su
-freeReg REG_Su   = fastBool False
-# endif
-# ifdef REG_SpLim
-freeReg REG_SpLim = fastBool False
-# endif
-# ifdef REG_Hp
-freeReg REG_Hp   = fastBool False
-# endif
-# ifdef REG_HpLim
-freeReg REG_HpLim = fastBool False
-# endif
-freeReg _               = fastBool True
+freeReg 30 = False
+{- TODO: reserve r13 on 64 bit systems only and r30 on 32 bit respectively.
+   For now we use r30 on 64 bit and r13 on 32 bit as a temporary register
+   in stack handling code. See compiler/nativeGen/PPC/Instr.hs.
 
-#elif MACHREGS_sparc
+   Later we might want to reserve r13 and r30 only where it is required.
+   Then use r12 as temporary register, which is also what the C ABI does.
+-}
+
+# if defined(REG_Base)
+freeReg REG_Base  = False
+# endif
+# if defined(REG_Sp)
+freeReg REG_Sp    = False
+# endif
+# if defined(REG_SpLim)
+freeReg REG_SpLim = False
+# endif
+# if defined(REG_Hp)
+freeReg REG_Hp    = False
+# endif
+# if defined(REG_HpLim)
+freeReg REG_HpLim = False
+# endif
+freeReg _ = True
+
+#elif defined(MACHREGS_sparc)
 
 -- SPARC regs used by the OS / ABI
 -- %g0(r0) is always zero
-freeReg g0  = fastBool False
+freeReg g0  = False
 
 -- %g5(r5) - %g7(r7)
 --  are reserved for the OS
-freeReg g5  = fastBool False
-freeReg g6  = fastBool False
-freeReg g7  = fastBool False
+freeReg g5  = False
+freeReg g6  = False
+freeReg g7  = False
 
 -- %o6(r14)
 --  is the C stack pointer
-freeReg o6  = fastBool False
+freeReg o6  = False
 
 -- %o7(r15)
 --  holds the C return address
-freeReg o7  = fastBool False
+freeReg o7  = False
 
 -- %i6(r30)
 --  is the C frame pointer
-freeReg i6  = fastBool False
+freeReg i6  = False
 
 -- %i7(r31)
 --  is used for C return addresses
-freeReg i7  = fastBool False
+freeReg i7  = False
 
 -- %f0(r32) - %f1(r32)
 --  are C floating point return regs
-freeReg f0  = fastBool False
-freeReg f1  = fastBool False
+freeReg f0  = False
+freeReg f1  = False
 
 {-
 freeReg regNo
@@ -1009,116 +987,112 @@ freeReg regNo
     | regNo >= f0
     , regNo <  NCG_FirstFloatReg
     , regNo `mod` 2 /= 0
-    = fastBool False
+    = False
 -}
 
-# ifdef REG_Base
-freeReg REG_Base = fastBool False
+# if defined(REG_Base)
+freeReg REG_Base  = False
 # endif
-# ifdef REG_R1
-freeReg REG_R1  = fastBool False
+# if defined(REG_R1)
+freeReg REG_R1    = False
 # endif
-# ifdef REG_R2
-freeReg REG_R2  = fastBool False
+# if defined(REG_R2)
+freeReg REG_R2    = False
 # endif
-# ifdef REG_R3
-freeReg REG_R3  = fastBool False
+# if defined(REG_R3)
+freeReg REG_R3    = False
 # endif
-# ifdef REG_R4
-freeReg REG_R4  = fastBool False
+# if defined(REG_R4)
+freeReg REG_R4    = False
 # endif
-# ifdef REG_R5
-freeReg REG_R5  = fastBool False
+# if defined(REG_R5)
+freeReg REG_R5    = False
 # endif
-# ifdef REG_R6
-freeReg REG_R6  = fastBool False
+# if defined(REG_R6)
+freeReg REG_R6    = False
 # endif
-# ifdef REG_R7
-freeReg REG_R7  = fastBool False
+# if defined(REG_R7)
+freeReg REG_R7    = False
 # endif
-# ifdef REG_R8
-freeReg REG_R8  = fastBool False
+# if defined(REG_R8)
+freeReg REG_R8    = False
 # endif
-# ifdef REG_R9
-freeReg REG_R9  = fastBool False
+# if defined(REG_R9)
+freeReg REG_R9    = False
 # endif
-# ifdef REG_R10
-freeReg REG_R10 = fastBool False
+# if defined(REG_R10)
+freeReg REG_R10   = False
 # endif
-# ifdef REG_F1
-freeReg REG_F1  = fastBool False
+# if defined(REG_F1)
+freeReg REG_F1    = False
 # endif
-# ifdef REG_F2
-freeReg REG_F2  = fastBool False
+# if defined(REG_F2)
+freeReg REG_F2    = False
 # endif
-# ifdef REG_F3
-freeReg REG_F3  = fastBool False
+# if defined(REG_F3)
+freeReg REG_F3    = False
 # endif
-# ifdef REG_F4
-freeReg REG_F4  = fastBool False
+# if defined(REG_F4)
+freeReg REG_F4    = False
 # endif
-# ifdef REG_F5
-freeReg REG_F5  = fastBool False
+# if defined(REG_F5)
+freeReg REG_F5    = False
 # endif
-# ifdef REG_F6
-freeReg REG_F6  = fastBool False
+# if defined(REG_F6)
+freeReg REG_F6    = False
 # endif
-# ifdef REG_D1
-freeReg REG_D1  = fastBool False
+# if defined(REG_D1)
+freeReg REG_D1    = False
 # endif
-# ifdef REG_D1_2
-freeReg REG_D1_2 = fastBool False
+# if defined(REG_D1_2)
+freeReg REG_D1_2  = False
 # endif
-# ifdef REG_D2
-freeReg REG_D2  = fastBool False
+# if defined(REG_D2)
+freeReg REG_D2    = False
 # endif
-# ifdef REG_D2_2
-freeReg REG_D2_2 = fastBool False
+# if defined(REG_D2_2)
+freeReg REG_D2_2  = False
 # endif
-# ifdef REG_D3
-freeReg REG_D3  = fastBool False
+# if defined(REG_D3)
+freeReg REG_D3    = False
 # endif
-# ifdef REG_D3_2
-freeReg REG_D3_2 = fastBool False
+# if defined(REG_D3_2)
+freeReg REG_D3_2  = False
 # endif
-# ifdef REG_D4
-freeReg REG_D4  = fastBool False
+# if defined(REG_D4)
+freeReg REG_D4    = False
 # endif
-# ifdef REG_D4_2
-freeReg REG_D4_2 = fastBool False
+# if defined(REG_D4_2)
+freeReg REG_D4_2  = False
 # endif
-# ifdef REG_D5
-freeReg REG_D5  = fastBool False
+# if defined(REG_D5)
+freeReg REG_D5    = False
 # endif
-# ifdef REG_D5_2
-freeReg REG_D5_2 = fastBool False
+# if defined(REG_D5_2)
+freeReg REG_D5_2  = False
 # endif
-# ifdef REG_D6
-freeReg REG_D6  = fastBool False
+# if defined(REG_D6)
+freeReg REG_D6    = False
 # endif
-# ifdef REG_D6_2
-freeReg REG_D6_2 = fastBool False
+# if defined(REG_D6_2)
+freeReg REG_D6_2  = False
 # endif
-# ifdef REG_Sp
-freeReg REG_Sp  = fastBool False
+# if defined(REG_Sp)
+freeReg REG_Sp    = False
 # endif
-# ifdef REG_Su
-freeReg REG_Su  = fastBool False
+# if defined(REG_SpLim)
+freeReg REG_SpLim = False
 # endif
-# ifdef REG_SpLim
-freeReg REG_SpLim = fastBool False
+# if defined(REG_Hp)
+freeReg REG_Hp    = False
 # endif
-# ifdef REG_Hp
-freeReg REG_Hp  = fastBool False
+# if defined(REG_HpLim)
+freeReg REG_HpLim = False
 # endif
-# ifdef REG_HpLim
-freeReg REG_HpLim = fastBool False
-# endif
-freeReg _   = fastBool True
+freeReg _ = True
 
 #else
 
 freeReg = panic "freeReg not defined for this platform"
 
 #endif
-
