@@ -523,10 +523,18 @@ stat_endGC (Capability *cap, gc_thread *initiating_gct, W_ live, W_ copied, W_ s
             initiating_gct->gc_start_elapsed - initiating_gct->gc_sync_start_elapsed;
         stats.gc.elapsed_ns = current_elapsed - initiating_gct->gc_start_elapsed;
         stats.gc.cpu_ns = 0;
-        for (unsigned int i=0; i < par_n_threads; i++) {
-            gc_thread *gct = gc_threads[i];
-            ASSERT(gct->gc_end_cpu >= gct->gc_start_cpu);
-            stats.gc.cpu_ns += gct->gc_end_cpu - gct->gc_start_cpu;
+        // see Note [n_gc_threads]
+        // par_n_threads is set to n_gc_threads at the single callsite of this
+        // function
+        if (par_n_threads == 1) {
+            ASSERT(initiating_gct->gc_end_cpu >= initiating_gct->gc_start_cpu);
+            stats.gc.cpu_ns += initiating_gct->gc_end_cpu - initiating_gct->gc_start_cpu;
+        } else {
+            for (unsigned int i=0; i < par_n_threads; i++) {
+                gc_thread *gct = gc_threads[i];
+                ASSERT(gct->gc_end_cpu >= gct->gc_start_cpu);
+                stats.gc.cpu_ns += gct->gc_end_cpu - gct->gc_start_cpu;
+            }
         }
     }
     // -------------------------------------------------
@@ -1387,7 +1395,7 @@ stat_exitReport (void)
                          * BLOCKS_PER_MBLOCK
                          * BLOCK_SIZE_W
                          - hw_alloc_blocks * BLOCK_SIZE_W)
-                / (uint64_t)sizeof(W_);
+                * (uint64_t)sizeof(W_);
 
             sum.average_bytes_used = stats.major_gcs == 0 ? 0 :
                  stats.cumulative_live_bytes/stats.major_gcs,
